////////////////////////////////////////////////////////////////////////
// Class:       AdcThresholdRoiFinder
// Plugin Type: producer (art v3_05_01)
// File:        AdcThresholdRoiFinder_module.cc
//
// Generated at Thu Jul  2 15:42:00 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
// About:  ROI finder based on Adc threshold.
// Reference: Refer David Adams's similiar work on DUNE
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib_except/exception.h"
#include "cetlib/search_path.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"

// larsoft includes
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardata/ArtDataHelper/WireCreator.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcore/Geometry/Geometry.h"

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

#include <memory>

using namespace std;


class AdcThresholdRoiFinder;


class AdcThresholdRoiFinder : public art::EDProducer {
public:
  explicit AdcThresholdRoiFinder(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  AdcThresholdRoiFinder(AdcThresholdRoiFinder const&) = delete;
  AdcThresholdRoiFinder(AdcThresholdRoiFinder&&) = delete;
  AdcThresholdRoiFinder& operator=(AdcThresholdRoiFinder const&) = delete;
  AdcThresholdRoiFinder& operator=(AdcThresholdRoiFinder&&) = delete;

  // Required functions.
  void produce(art::Event& e) override;

private:

  // configuration from fhicl file
  std::string fRawDigitLabel;
  std::string fWireProducerLabel;
  std::vector<float> fThreshold;
  bool fRebinWaveform;
  int fRebinFactor;
  int fBinsBefore;
  int fBinsAfter;
  bool fFlagPositive;
  bool fFlagNegtive;

  int fNticks;

};


AdcThresholdRoiFinder::AdcThresholdRoiFinder(fhicl::ParameterSet const& p)
  : EDProducer{p}  // ,
  // More initializers here.
{
  fRawDigitLabel     =  p.get<std::string>("RawDigitLabel");
  fWireProducerLabel =  p.get<std::string>("WireProducerLabel");
  fThreshold               =  p.get<std::vector<float>>("Threshold");
  fRebinWaveform           =  p.get<bool>("RebinWaveform");
  fRebinFactor             =  p.get<int>("RebinFactor");
  fBinsBefore              =  p.get<int>("BinsBefore");
  fBinsAfter               =  p.get<int>("BinsAfter");
  fFlagPositive            =  p.get<bool>("FlagPositive");
  fFlagNegtive             =  p.get<bool>("FlagNegtive");

  if (fRawDigitLabel.empty() && fWireProducerLabel.empty()){
    throw cet::exception("AdcThresholdRoiFinder") << "Both RawDigitLabel and WireProducerLabel are empty";
  }

  if ((!fRawDigitLabel.empty()) && (!fWireProducerLabel.empty())){
    throw cet::exception("AdcThresholdRoiFinder") << "Only one of RawDigitLabel and WireProducerLabel should be set";
  }

  // DetectorPropertiesService
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob();
  fNticks = detProp.ReadOutWindowSize(); // number of clock ticks per event
  //std::cout << "Number of clock ticks per event: " << fNticks << std::endl;

  // Call appropriate produces<>() functions here.
  // Call appropriate consumes<>() for any products to be retrieved by this module.
  produces< std::vector<recob::Wire> >();

}

void AdcThresholdRoiFinder::produce(art::Event& e)
{
  
  auto const* geo = lar::providerFrom<geo::Geometry>();

  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, "", rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }
  
  //cout << "rawdigitlist.size(): " << rawdigitlist.size() << endl;

  // get Wire
  art::Handle< std::vector<recob::Wire> > wireListHandle;
  std::vector<art::Ptr<recob::Wire> > wirelist;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wirelist, wireListHandle);
  }

  //cout << "wirelist.size(): " << wirelist.size() << endl;

  // make a collection of Wires and an association set
  std::unique_ptr<std::vector<recob::Wire> > wirecol(new std::vector<recob::Wire>);

  // loop over wires
  // use one of rawdigitlist and  wirelist that is not empty()
  for (unsigned int ich=0; ich < (rawdigitlist.empty()?wirelist.size():rawdigitlist.size()); ++ich) {

    std::vector<float> adcvec(fNticks);

    auto view = geo->View(ich);

    if (!rawdigitlist.empty()) {
      const auto & digitVec = rawdigitlist[ich];
  
      view =  geo->View(digitVec->Channel());

      std::vector<short> rawadc(fNticks);
      raw::Uncompress(digitVec->ADCs(), rawadc, digitVec->GetPedestal(), digitVec->Compression());

      for (size_t j=0; j<rawadc.size(); ++j) {
        adcvec[j] = rawadc[j] - digitVec->GetPedestal(); 
      }
    }
    else if (!wirelist.empty()) {
      const auto & wire = wirelist[ich];
      const auto & signal = wire->Signal();
      
      view = wirelist[ich]->View();

      for (size_t j=0; j<adcvec.size(); ++j) {
        adcvec[j] = signal[j];
      }
    }
    
    /*
    if (ich == 120) {
      for (size_t j=0; j<adcvec.size(); ++j) {
        cout << adcvec[j] << endl;
      }
    }
    */

    recob::Wire::RegionsOfInterest_t rois(fNticks);

    std::vector<float> sigs;
    int lastsignaltick = -1;
    int roistart = -1;

    int jUnknown = 0; // first sample which is not known to be inside or outside a ROI
    
    // there are two options for rebinning the waveform, for a rebin factor of 2: a) 2048->1024 with unit of 2 ticks; b) 2048->2048 with unit of 1 tick
    // we choose the later one here
    for (size_t j=0; j<(fRebinWaveform ? adcvec.size()/fRebinFactor : adcvec.size()); ++j) {

      bool isroi;
      if (fRebinWaveform) {
        float adcrebin = adcvec[fRebinFactor*j];
        for (int jkk=1; jkk<fRebinFactor; jkk++) {
          adcrebin += adcvec[fRebinFactor*j+jkk];
        }
        isroi  = ( (fFlagPositive && adcrebin > fThreshold[view]) || (fFlagNegtive && adcrebin < -fThreshold[view]) );
      }
      else {
        isroi  = ( (fFlagPositive && adcvec[j] > fThreshold[view]) || (fFlagNegtive && adcvec[j] < -fThreshold[view]) );
      }
      
      if (fRebinWaveform) {
        if (isroi && int(j*fRebinFactor) > lastsignaltick) {

          roistart = int(j*fRebinFactor)>fBinsBefore ? fRebinFactor*j-fBinsBefore : 0;
          if (roistart < jUnknown) roistart = jUnknown;

          lastsignaltick = fRebinFactor*j+fRebinFactor-1 + fBinsAfter;
          if (lastsignaltick > fNticks-1) lastsignaltick = fNticks-1;

          for (int jk=roistart; jk<=lastsignaltick; jk++) {
            sigs.push_back(adcvec[jk]);
            //cout << "j: " << j << "; jk: " << jk << "; isroi: " << isroi << "; adc" << adcvec[jk] << endl;
          }

          jUnknown = lastsignaltick+1;

          rois.add_range(roistart, std::move(sigs));
          sigs.clear();
        }
      }
      else {
        if (isroi && int(j) > lastsignaltick) {

          roistart = int(j)>fBinsBefore ? j-fBinsBefore : 0;
          if (roistart < jUnknown) roistart = jUnknown;

          lastsignaltick = j + fBinsAfter;
          if (lastsignaltick > fNticks-1) lastsignaltick = fNticks-1;

          for (int jk=roistart; jk<=lastsignaltick; jk++) {
            sigs.push_back(adcvec[jk]);
          }

          jUnknown = lastsignaltick+1;

          rois.add_range(roistart, std::move(sigs));
          sigs.clear();
        }
      }

    } // loop adcvec j

    if (!rawdigitlist.empty()) {
      wirecol->emplace_back(recob::Wire(rois, rawdigitlist[ich]->Channel(), geo->View(rawdigitlist[ich]->Channel())));
    }
    else if (!wirelist.empty()) {
      wirecol->emplace_back(recob::Wire(rois, wirelist[ich]->Channel(), wirelist[ich]->View()));
    }
  
  } //loop wire ich


  if (wirecol->size() == 0) {
    mf::LogWarning("CalWireArgoneut") << "No wires made for this event.";
  }

  e.put(std::move(wirecol));
}

DEFINE_ART_MODULE(AdcThresholdRoiFinder)
